from typing import Annotated
from fastmcp import FastMCP
from pathlib import Path

from iolanta.cli.main import render_and_return

mcp = FastMCP("Iolanta MCP Server")


@mcp.tool()
def render_uri(
    uri: Annotated[str, 'URL, or file system path, to render'],
    as_format: Annotated[str, 'Format to render as. Examples: `labeled-triple-set`, `mermaid`']
) -> str:
    """Render a URI."""
    result = render_and_return(uri, as_format)
    return str(result)


@mcp.prompt(description="How to author Linked Data with Iolanta")
def ld_authoring_rules() -> str:
    """How to author Linked Data with Iolanta."""
    rules_path = Path(__file__).parent / 'prompts' / 'rules.md'
    return rules_path.read_text()


@mcp.prompt(description="How to author nanopublication assertions with Iolanta")
def nanopublication_assertion_authoring_rules() -> str:
    """How to author nanopublication assertions with Iolanta."""
    rules_path = Path(__file__).parent / 'prompts' / 'nanopublication_assertion_authoring_rules.md'
    return rules_path.read_text()


def app():
    mcp.run()


if __name__ == "__main__":
    app()



import locale
import logging
import sys
from pathlib import Path
from typing import Annotated

import loguru
import platformdirs
from documented import DocumentedError
from rdflib import Literal, URIRef
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table
from typer import Argument, Exit, Option, Typer, Context
from yarl import URL

from iolanta.cli.models import LogLevel
from iolanta.iolanta import Iolanta
from iolanta.models import NotLiteralNode

DEFAULT_LANGUAGE = locale.getlocale()[0].split('_')[0]


console = Console()


app = Typer(no_args_is_help=True)


def string_to_node(name: str) -> NotLiteralNode:
    """
    Parse a string into a node identifier.

    String might be:
      * a URL,
      * or a local disk path.
    """
    url = URL(name)
    if url.scheme:
        return URIRef(name)

    path = Path(name).absolute()
    return URIRef(f'file://{path}')


def decode_datatype(datatype: str) -> URIRef:
    if datatype.startswith('http'):
        return URIRef(datatype)

    return URIRef(f'https://iolanta.tech/datatypes/{datatype}')


def render_and_return(
    url: str,
    as_datatype: str,
    language: str = DEFAULT_LANGUAGE,
    log_level: LogLevel = LogLevel.ERROR,
):
    """Render a given URL."""
    level = {
        LogLevel.DEBUG: logging.DEBUG,
        LogLevel.INFO: logging.INFO,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.ERROR: logging.ERROR,
    }[log_level]

    log_file_path = platformdirs.user_log_path(
        'iolanta',
        ensure_exists=True,
    ) / 'iolanta.log'

    # Get the level name first
    level_name = {
        logging.DEBUG: 'DEBUG',
        logging.INFO: 'INFO', 
        logging.WARNING: 'WARNING',
        logging.ERROR: 'ERROR',
    }[level]
    
    # Configure global loguru logger BEFORE creating Iolanta instance
    loguru.logger.remove()
    loguru.logger.add(
        log_file_path,
        level=level_name,
        format='{time} {level} {message}',
        enqueue=True,
    )
    loguru.logger.add(
        sys.stderr,
        level=level_name,
        format='{time} | {level:<8} | {name}:{function}:{line} - {message}',
    )
    loguru.logger.level(level_name)
    
    # Use the global logger
    logger = loguru.logger
    
    node_url = URL(url)
    if node_url.scheme and node_url.scheme != 'file':
        node = URIRef(url)

        iolanta: Iolanta = Iolanta(
            language=Literal(language),
            logger=logger,
        )
        
    else:
        path = Path(node_url.path).absolute()
        node = URIRef(f'file://{path}')
        iolanta: Iolanta = Iolanta(
            language=Literal(language),
            logger=logger,
            project_root=path,
        )

    return iolanta.render(
        node=URIRef(node),
        as_datatype=decode_datatype(as_datatype),
    )


@app.command(name='render')
def render_command(   # noqa: WPS231, WPS238, WPS210, C901
    url: Annotated[str, Argument()],
    as_datatype: Annotated[
        str, Option(
            '--as',
        ),
    ] = 'https://iolanta.tech/cli/interactive',
    language: Annotated[
        str, Option(
            help='Data language to prefer.',
        ),
    ] = DEFAULT_LANGUAGE,
    log_level: LogLevel = LogLevel.ERROR,
):
    """Render a given URL."""
    try:
        renderable = render_and_return(url, as_datatype, language, log_level)
    except DocumentedError as documented_error:
        level = {
            LogLevel.DEBUG: logging.DEBUG,
            LogLevel.INFO: logging.INFO,
            LogLevel.WARNING: logging.WARNING,
            LogLevel.ERROR: logging.ERROR,
        }[log_level]
        
        if level in {logging.DEBUG, logging.INFO}:
            raise

        console.print(
            Markdown(
                str(documented_error),
                justify='left',
            ),
        )
        raise Exit(1)

    except Exception as err:
        level = {
            LogLevel.DEBUG: logging.DEBUG,
            LogLevel.INFO: logging.INFO,
            LogLevel.WARNING: logging.WARNING,
            LogLevel.ERROR: logging.ERROR,
        }[log_level]
        
        if level in {logging.DEBUG, logging.INFO}:
            raise

        console.print(str(err))
        raise Exit(1)
    else:
        # FIXME: An intermediary Literal can be used to dispatch rendering.
        match renderable:
            case Table() as table:
                console.print(table)

            case unknown:
                print(unknown)



import functools
import itertools
from typing import ClassVar, Iterable

import funcy
from rich.console import RenderResult
from rich.text import Text
from textual.binding import Binding, BindingType
from textual.containers import Vertical  # removed unused VerticalScroll (F401)
from textual.reactive import Reactive
from textual.widget import Widget
from textual.widgets import Label, ListItem, ListView

from iolanta.facets.facet import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.facets.textual_default.triple_uri_ref import TripleURIRef
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES, RDF

INSTANCE_RENDER_RADIUS = 50


class InstanceItem(ListItem):
    """An item in class instances list."""

    renderable: Reactive[RenderResult | None](None, init=True, layout=True)

    def __init__(
        self,
        node: NotLiteralNode,
        parent_class: NotLiteralNode,
    ):
        """Specify the node, its class, and that we are not rendered yet."""
        self.node = node
        self.parent_class = parent_class
        self.renderable = None
        super().__init__()

    def render(self) -> RenderResult:
        """
        Render this class instance.

        Render either the result or the raw node.

        # FIXME Calculate QName at least. Or CURIE? Or whatsit?
        """
        qname = self.app.iolanta.node_as_qname(self.node)
        return self.renderable or Text(
            f'⏳ {qname}',
            style='#696969',
        )

    def update(self, renderable: RenderResult):
        """
        Assign the render result.

        A separate method is needed for this because we call it from a thread.
        """
        self.renderable = renderable
        self.refresh()


class InstancesList(ListView):   # noqa: WPS214
    """Instances list."""

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding('enter', 'goto', 'Goto'),
        Binding('p', 'provenance', 'Provenan©e'),
    ]

    DEFAULT_CSS = """
    InstancesList {
        height: 1fr;
        layout: vertical;
        overflow-y: auto;
    }
    """

    FIRST_CHUNK_SIZE = 15
    DEFAULT_CHUNK_SIZE = 10

    def __init__(
        self,
        instances: Iterable[NotLiteralNode],
        parent_class: NotLiteralNode,
    ):
        """Specify the instances to render and their class."""
        self.instances = instances
        self.parent_class = parent_class
        super().__init__()

    def render_newly_added_instances(self, instance_items: list[InstanceItem]):
        """
        Render each of the given instance items.

        Must be called in a worker.
        """
        for instance_item in instance_items:
            self.app.call_from_thread(
                instance_item.update,
                self.app.iolanta.render(
                    instance_item.node,
                    as_datatype=DATATYPES.title,
                ),
            )

    def stream_instance_items_chunk(
        self,
        count: int | None = None,
    ) -> Iterable[InstanceItem]:
        """Return a chunk of unique class instances."""
        chunk = itertools.islice(
            self.instances,
            count or self.DEFAULT_CHUNK_SIZE,
        )
        for instance in chunk:  # noqa: WPS526
            yield InstanceItem(
                node=instance,
                parent_class=self.parent_class,
            )

    def action_goto(self):
        """Navigate."""
        self.app.action_goto(self.highlighted_child.node)

    def action_provenance(self):
        """Navigate to provenance for the property value."""
        self.app.action_goto(
            TripleURIRef.from_triple(
                subject=self.highlighted_child.node,
                predicate=RDF.type,
                object_=self.parent_class,
            ),
        )

    def compose(self):
        """Fill in the instances."""
        for instance in self.instances:   # noqa: WPS526
            yield InstanceItem(
                node=instance,
                parent_class=self.parent_class,
            )

    def render_instances(self):
        """Render all instances."""
        for instance_item in self.children:
            self.app.call_from_thread(
                instance_item.update,
                self.app.iolanta.render(
                    instance_item.node,
                    as_datatype=DATATYPES.title,
                ),
            )

    def on_mount(self):
        """Render the first chunk of instances."""
        self.run_worker(
            self.render_instances,
            thread=True,
        )


class Bottom(Label):
    """Label below the instances list."""

    DEFAULT_CSS = """
    Bottom {
        padding-top: 1;
        padding-bottom: 1;
        dock: bottom;
    }
    """


class InstancesBody(Vertical):
    """Container for instances list and accompanying bells and whistles."""

    DEFAULT_CSS = """
    InstancesBody {
        height: auto;
        max-height: 100%;
    }
    """


class Class(Facet[Widget]):
    """Render instances of a class."""

    def stream_instances(self) -> Iterable[NotLiteralNode]:
        """
        Query and stream class instances lazily.

        The operation of rendering an instance is not pure: it might cause us
        to retrieve more data and load said data into the graph. That's because
        we do multiple query attempts.

        We have to stop if a subsequent attempt returns no results. That's why
        we can't use `funcy.distinct()` or something similar.
        """
        return set(
            funcy.pluck(
                'instance',
                self.stored_query('instances.sparql', this=self.this),
            ),
        )

    def show(self) -> Widget:
        """Render the instances list."""
        return InstancesBody(
            PageTitle(self.this),
            InstancesList(
                instances=list(self.stream_instances()),
                parent_class=self.this,
            ),
        )



import functools
from dataclasses import dataclass, field
from pathlib import Path
from typing import Annotated

import rdflib
from rdflib.plugins.sparql.sparql import Query
from rdflib.query import Result


@dataclass
class SPARQLSpace:
    """SPARQL interface to the Web of Data."""

    graph: rdflib.Dataset = field(
        default_factory=functools.partial(
            rdflib.Dataset,
            default_union=True,
        ),
    )
    path: Annotated[
        Path | None,
        'Local directory to load into the graph.',
    ] = None

    def query(self, query: str | Query, **bindings) -> Result:
        """Execute a SPARQL query."""
        return self.graph.query(
            query_object=query,
            processor='sparqlspace',
            initBindings=bindings,
        )



from rdflib import URIRef
from rdflib.term import Node
from yarl import URL

from iolanta.models import NotLiteralNode, Triple


class TripleURIRef(URIRef):
    """URN serialization of an RDF triple."""

    @classmethod
    def from_triple(
        cls,
        subject: NotLiteralNode,
        predicate: NotLiteralNode,
        object_: Node,
    ) -> 'TripleURIRef':
        """
        Construct from triple.

        TODO Add special query arguments to conform to RDF standard:
          * subject_bnode
          * predicate_bnode
          * object_bnode
          * object_datatype
          * object_language

        TODO Standardize this?
        """
        iri = URL.build(
            scheme='urn:rdf',
            query={
                'subject': subject,
                'predicate': predicate,
                'object': object_,
            },
        )
        return TripleURIRef(str(iri))

    def as_triple(self) -> Triple:
        """
        Deserialize into a triple.

        TODO support special query arguments described above.
        """
        url = URL(self)
        return Triple(
            subject=URIRef(url.query['subject']),
            predicate=URIRef(url.query['predicate']),
            object=URIRef(url.query['object']),
        )



from textual.containers import ScrollableContainer

from iolanta.facets.textual_browser.models import FlipOption
from iolanta.models import NotLiteralNode


class Page(ScrollableContainer):
    """Page in Iolanta browser."""

    def __init__(
        self,
        renderable,
        iri: NotLiteralNode,
        page_id: str,
        flip_options: list[FlipOption],
    ):
        """Initialize the page and set bindings."""
        super().__init__(renderable, id=page_id)
        for number, flip_option in enumerate(flip_options, start=1):
            self._bindings.bind(
                keys=str(number),
                description=flip_option.title,
                action=(
                    f"app.goto('{iri}', '{flip_option.facet_iri}')"
                ),
            )



import functools
from pathlib import Path
from typing import Iterable

import boltons
import cachetools
import funcy
from boltons.cacheutils import cached, cachedmethod
from pydantic import AnyUrl
from rdflib import BNode, Literal, Node, URIRef

from iolanta import Facet
from iolanta.mermaid.models import (
    Diagram,
    MermaidBlankNode,
    MermaidEdge,
    MermaidLiteral,
    MermaidScalar,
    MermaidSubgraph,
    MermaidURINode,
)
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES


def filter_edges(edges: Iterable[MermaidEdge]) -> Iterable[MermaidEdge]:
    for edge in edges:
        if isinstance(edge.target, MermaidLiteral) and edge.source.title == edge.target.title:
            continue

        yield edge


def filter_nodes(edges: Iterable[MermaidEdge], except_uris: Iterable[NotLiteralNode]) -> Iterable[MermaidURINode | MermaidLiteral | MermaidBlankNode]:
    nodes = [
        node
        for edge in edges
        for node in edge.nodes
    ]

    literals_in_edges = {edge.target for edge in edges if isinstance(edge.target, MermaidLiteral)}
    for node in nodes:
        if isinstance(node, MermaidLiteral) and node not in literals_in_edges:
            continue

        if isinstance(node, MermaidURINode) and node.uri in except_uris:
            continue

        if isinstance(node, MermaidBlankNode) and node.node in except_uris:
            continue

        if isinstance(node, MermaidSubgraph):
            continue

        yield node


class Mermaid(Facet[str]):
    """Mermaid diagram."""

    META = Path(__file__).parent / 'mermaid.yamlld'

    def as_mermaid(self, node: Node):
        match node:
            case URIRef() as uri:
                if uri in self.subgraph_uris:
                    return MermaidSubgraph(children=[], uri=uri, title=self.render(uri, as_datatype=DATATYPES.title))

                return MermaidURINode(uri=uri, url=AnyUrl(uri), title=self.render(uri, as_datatype=DATATYPES.title))
            case Literal() as literal:
                return MermaidLiteral(literal=literal)
            case BNode() as bnode:
                return MermaidBlankNode(node=bnode, title=self.render(bnode, as_datatype=DATATYPES.title))
            case unknown:
                unknown_type = type(unknown)
                raise ValueError(f'Unknown something: {unknown} ({unknown_type})')

    def construct_mermaid_for_graph(self, graph: URIRef) -> Iterable[MermaidScalar]:
        """Render graph as mermaid."""
        rows = self.stored_query('graph.sparql', this=graph)
        edges = [
            MermaidEdge(
                source=self.as_mermaid(row['s']),
                target=self.as_mermaid(row['o']),
                title=self.render(row['p'], as_datatype=DATATYPES.title),
                predicate=row['p'],
            ) for row in rows
        ]

        edges = list(filter_edges(edges))
        nodes = list(
            filter_nodes(
                edges=edges,
                except_uris=self.subgraph_uris,
            ),
        )

        return *nodes, *edges

    @functools.cached_property
    def subgraph_uris(self) -> set[NotLiteralNode]:
        return set(
            funcy.pluck(
                'subgraph',
                self.stored_query('subgraphs.sparql', this=self.this),
            ),
        )

    def construct_mermaid_subgraphs(self) -> Iterable[MermaidSubgraph]:
        for subgraph_uri in self.subgraph_uris:
            children = list(self.construct_mermaid_for_graph(subgraph_uri))
            if children:
                title = self.render(subgraph_uri, as_datatype=DATATYPES.title)
                yield MermaidSubgraph(
                    children=children,
                    uri=subgraph_uri,
                    title=title,
                )

    def show(self) -> str:
        """Render mermaid diagram."""
        direct_children = self.construct_mermaid_for_graph(self.this)
        subgraphs = self.construct_mermaid_subgraphs()
        return str(Diagram(children=[*direct_children, *subgraphs]))



from abc import abstractmethod
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Optional, TypedDict

import funcy
from rdflib import Literal

from iolanta.facets.facet import Facet, FacetOutput


@dataclass
class Description:
    """A few properties of the object we use as row heading."""

    label: Optional[Literal] = None
    symbol: Optional[Literal] = None
    url: Optional[Literal] = None
    comment: Optional[Literal] = None


class DefaultMixin(Facet[FacetOutput]):
    stored_queries_path = Path(__file__).parent / 'sparql'

    @cached_property
    def description(self) -> Description:
        return Description(
            **funcy.first(
                self.stored_query('default.sparql', iri=self.this),
            ),
        )

    def show(self) -> FacetOutput:
        """Render the column."""
        if self.description.url:
            return self.render_link()

        return self.render_label()

    @abstractmethod
    def render_link(self) -> FacetOutput:
        """Render clickable link."""

    def render_label(self) -> str:
        if not (label := self.description.label):
            label = self.render_fallback()

        if isinstance(label, Literal):
            label = label.value

        if symbol := self.description.symbol:
            rendered_symbol = self.render(
                symbol,
                as_datatype=self.as_datatype,
            )
            label = f'{rendered_symbol} {label}'

        return label

    def render_fallback(self) -> str:
        string_iri = str(self.this)

        if string_iri.startswith('local:'):
            string_iri = string_iri.removeprefix(
                'local:',
            ).replace(
                '_', ' ',
            ).replace(
                '-', ' ',
            ).capitalize()

        return string_iri



from rdflib import Literal

from iolanta.facets.errors import NotALiteral
from iolanta.facets.facet import Facet


class BoolLiteral(Facet):
    """Render bool values."""

    def show(self):
        """Render as icon."""
        if not isinstance(self.this, Literal):
            raise NotALiteral(
                node=self.this,
            )

        return '✔️' if self.this.value else '❌'



from textual.containers import VerticalScroll
from textual.widget import Widget

from iolanta.facets.facet import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.models import Triple
from iolanta.namespaces import DATATYPES


class GraphFacet(Facet[Widget]):
    """Display triples in a graph."""

    def show(self) -> Widget:
        """Show the widget."""
        triples = [
            Triple(triple['subject'], triple['predicate'], triple['object'])
            for triple in self.stored_query('triples.sparql', graph=self.this)
        ]

        triple_count = len(triples)

        triples_view = self.iolanta.render(
            self.this,
            as_datatype=DATATYPES['textual-graph-triples'],
        )

        return VerticalScroll(
            PageTitle(self.this, extra=f'({triple_count} triples)'),
            triples_view,
        )



import dataclasses
import hashlib
from types import MappingProxyType
from typing import Iterable, Optional
from urllib.parse import unquote

from documented import DocumentedError
from rdflib import BNode, Literal, URIRef
from rdflib.term import Node

from iolanta.errors import UnresolvedIRI
from iolanta.models import Quad
from iolanta.namespaces import IOLANTA, META

NORMALIZE_TERMS_MAP = MappingProxyType({
    URIRef(_url := 'http://www.w3.org/2002/07/owl'): URIRef(f'{_url}#'),
    URIRef(_url := 'http://www.w3.org/2000/01/rdf-schema'): URIRef(f'{_url}#'),
})


def parse_term(   # noqa: C901
    term,
    blank_node_prefix,
) -> Node:
    """Parse N-Quads term into a Quad."""
    if term is None:
        raise SpaceInProperty()

    term_type = term['type']
    term_value = term['value']

    if term_type == 'IRI':
        return URIRef(unquote(term_value))

    if term_type == 'literal':
        language = term.get('language')

        if datatype := term.get('datatype'):
            datatype = URIRef(datatype)

        if language and datatype:
            datatype = None

        return Literal(
            term_value,
            datatype=datatype,
            lang=language,
        )

    if term_type == 'blank node':
        return BNode(
            value=term_value.replace('_:', f'{blank_node_prefix}_'),
        )

    raise ValueError(f'Unknown term: {term}')


def construct_subgraph_name(subgraph_name: str, graph: URIRef) -> URIRef:
    """
    Construct a proper subgraph name URI from a base name and graph.

    If the subgraph name already starts with the graph URI, return it as is.
    Otherwise, append the name as a fragment to the graph URI.
    """
    if subgraph_name.startswith(str(graph)):
        return URIRef(subgraph_name)

    return URIRef(f'{graph}#{subgraph_name}')


def _parse_quads_per_subgraph(
    raw_quads,
    blank_node_prefix: str,
    graph: URIRef,
    subgraph: URIRef,
) -> Iterable[Quad]:
    for quad in raw_quads:
        try:
            yield Quad(
                subject=parse_term(quad['subject'], blank_node_prefix),
                predicate=parse_term(quad['predicate'], blank_node_prefix),
                object=parse_term(quad['object'], blank_node_prefix),
                graph=subgraph,
            )
        except SpaceInProperty as err:
            raise dataclasses.replace(
                err,
                iri=graph,
            )


def parse_quads(
    quads_document,
    graph: URIRef,
    blank_node_prefix: str = '',
) -> Iterable[Quad]:
    """Parse an N-Quads output into a Quads stream."""
    blank_node_prefix = hashlib.md5(  # noqa: S324
        blank_node_prefix.encode(),
    ).hexdigest()
    blank_node_prefix = f'_:{blank_node_prefix}'

    subgraph_names = {
        URIRef(subgraph_name): construct_subgraph_name(
            subgraph_name,
            graph=graph,
        )
        for subgraph_name in quads_document.keys()
        if subgraph_name != '@default'
    }
    subgraph_names[graph] = graph

    for subgraph, quads in quads_document.items():
        if subgraph == '@default':
            subgraph = graph   # noqa: WPS440

        else:
            subgraph = URIRef(subgraph)

            yield Quad(
                graph,
                IOLANTA['has-sub-graph'],
                subgraph_names[subgraph],
                META,
            )

        quads = _parse_quads_per_subgraph(
            quads,
            blank_node_prefix=blank_node_prefix,
            graph=subgraph,
            subgraph=subgraph_names[subgraph],
        )

        for quad in quads:   # noqa: WPS526
            yield quad.replace(
                subgraph_names | NORMALIZE_TERMS_MAP | {
                    # To enable nanopub rendering
                    URIRef('http://purl.org/nanopub/temp/np/'): graph,
                },
            ).normalize()


def raise_if_term_is_qname(term_value: str):
    """Raise an error if a QName is provided instead of a full IRI."""
    prefix, etc = term_value.split(':', 1)

    if etc.startswith('/'):
        return

    if prefix in {'local', 'templates', 'urn'}:
        return

    raise UnresolvedIRI(
        iri=term_value,
        prefix=prefix,
    )


@dataclasses.dataclass
class SpaceInProperty(DocumentedError):
    """
    Space in property.

    That impedes JSON-LD parsing.

    Please do not use spaces in property names in JSON or YAML data; use `title`
    or other methods instead.

    Document IRI: {self.iri}
    """

    iri: Optional[URIRef] = None



from typing import Type

from rdflib import URIRef
from yarl import URL

from iolanta import Facet
from iolanta.resolvers.base import Resolver


class SchemeDispatchResolver(Resolver):
    """
    A resolver that dispatches to other resolvers based on URI scheme.

    For example, 'pkg:' URIs are handled by PyPIResolver, while
    'python:' URIs are handled by PythonImportResolver.
    """

    def __init__(self, **resolver_by_scheme: Type[Resolver]):
        """
        Initialize with a mapping of URI schemes to resolver classes.

        Args:
            **resolver_by_scheme: Mapping of URI schemes to resolver classes.
                For example: python=PythonImportResolver.
        """
        self.resolver_by_scheme = resolver_by_scheme

    def resolve(self, uri: URIRef) -> Type[Facet]:
        """
        Find a resolver by IRI, dispatching to the appropriate scheme handler.

        Args:
            uri: The URI to resolve.

        Returns:
            Facet class resolved by the appropriate scheme handler.
        """
        url = URL(uri)
        resolver_class = self.resolver_by_scheme[url.scheme]
        resolver = resolver_class()
        return resolver.resolve(uri)



import enum
import hashlib
import re
import textwrap
import urllib.parse


def escape_label(label: str) -> str:
    """Escape a label to prevent Mermaid from interpreting URLs as markdown links."""
    # Remove https://, http://, and www. prefixes to prevent markdown link parsing
    return label.replace('https://', '').replace('http://', '').replace('www.', '')

from documented import Documented
from pydantic import AnyUrl, BaseModel
from rdflib import BNode, Literal, URIRef

from iolanta.models import NotLiteralNode


class Direction(enum.StrEnum):
    """Mermaid diagram direction."""

    TB = 'TB'
    LR = 'LR'


class MermaidURINode(Documented, BaseModel, arbitrary_types_allowed=True, frozen=True):
    """
    {self.id}{self.maybe_title}
    click {self.id} "{self.url}"
    """

    uri: URIRef
    url: AnyUrl
    title: str = ''

    @property
    def maybe_title(self):
        if not self.title:
            return ''
        # Escape URLs to prevent Mermaid from interpreting them as markdown links
        safe_title = escape_label(self.title)
        return f'("{safe_title}")'

    @property
    def id(self):
        return re.sub(r'[:\/\.#()]', '_', urllib.parse.unquote(str(self.url)).strip('/'))


class MermaidLiteral(Documented, BaseModel, arbitrary_types_allowed=True, frozen=True):
    """{self.id}[["{self.title}"]]"""

    literal: Literal

    @property
    def title(self) -> str:
        raw_title = str(self.literal) or 'EMPTY'
        # Replace quotes with safer characters for Mermaid
        return raw_title.replace('"', '"').replace("'", "'")

    @property
    def id(self) -> str:
        # Use the lexical form of the literal, not rdflib's .value (which may be empty for typed literals),
        # to ensure different texts get distinct node IDs in Mermaid.
        value_hash = hashlib.md5(str(self.literal).encode()).hexdigest()
        return f'Literal-{value_hash}'


class MermaidBlankNode(Documented, BaseModel, arbitrary_types_allowed=True):
    """{self.id}({self.escaped_title})"""

    node: BNode
    title: str

    @property
    def id(self) -> str:
        return self.node.replace('_:', '')
    
    @property
    def escaped_title(self) -> str:
        return self.title


class MermaidEdge(Documented, BaseModel, arbitrary_types_allowed=True):
    """
    {self.source.id} --- {self.id}(["{self.escaped_title}"])--> {self.target.id}
    click {self.id} "{self.predicate}"
    class {self.id} predicate
    """

    source: 'MermaidURINode | MermaidBlankNode | MermaidSubgraph'
    target: 'MermaidURINode | MermaidLiteral | MermaidBlankNode | MermaidSubgraph'
    predicate: URIRef
    title: str

    @property
    def id(self) -> str:
        return hashlib.md5(f'{self.source.id}{self.predicate}{self.target.id}'.encode()).hexdigest()

    @property
    def nodes(self):
        return [self.source, self.target]
    
    @property
    def escaped_title(self) -> str:
        # Escape URLs to prevent Mermaid from interpreting them as markdown links
        return escape_label(self.title)


MermaidScalar = MermaidLiteral | MermaidBlankNode | MermaidURINode | MermaidEdge


class MermaidSubgraph(Documented, BaseModel, arbitrary_types_allowed=True, frozen=True):
    """
    subgraph {self.id}["{self.escaped_title}"]
      direction {self.direction}
      {self.formatted_body}
    end
    """
    children: list[MermaidScalar]
    uri: NotLiteralNode
    title: str
    direction: Direction = Direction.LR

    @property
    def id(self):
        uri_hash = hashlib.md5(str(self.uri).encode()).hexdigest()
        return f'subgraph_{uri_hash}'
    
    @property
    def escaped_title(self) -> str:
        """Escape the subgraph title to prevent markdown link parsing."""
        return escape_label(self.title)

    @property
    def formatted_body(self):
        return textwrap.indent(
            '\n'.join(map(str, self.children)),
            prefix='  ',
        )


class Diagram(Documented, BaseModel):
    """
    graph {self.direction}
    {self.formatted_body}
      classDef predicate fill:transparent,stroke:transparent,stroke-width:0px;
    """

    children: list[MermaidScalar | MermaidSubgraph]
    direction: Direction = Direction.LR

    @property
    def formatted_body(self):
        return textwrap.indent(
            '\n'.join(map(str, self.children)),
            prefix='  ',
        )



from dataclasses import dataclass
from typing import Optional

from documented import DocumentedError
from rdflib.term import Node

from iolanta.models import LDContext


@dataclass
class InsufficientDataForRender(DocumentedError):
    """
    Insufficient data for rendering {self.node} — will try & download something.
    """

    node: Node
    iolanta: 'iolanta.Iolanta'

    @property
    def is_hopeless(self) -> bool:
        hopeless = self.node in self.iolanta.could_not_retrieve_nodes

        if hopeless:
            self.iolanta.logger.error(
                '%s could not be rendered, we could not retrieve describing '
                'data '
                'from the Web.',
                self.node,
            )

        return hopeless


@dataclass
class UnresolvedIRI(DocumentedError):
    """
    An unresolved IRI found.

        IRI: {self.iri}
        file: {self.file}
        prefix: {self.prefix}

    Perhaps you forgot to import appropriate context? For example:

    ```yaml
    "@context":
        - {self.prefix}: https://example.com/{self.prefix}/
    ```

    Context: {self.context}
    """

    iri: str
    prefix: str
    file: Optional[str] = None    # noqa: WPS110
    context: Optional[LDContext] = None



from textual.widget import Widget

from iolanta import Facet
from iolanta.facets.textual_nanopublication.nanopublication_widget import (
    NanopublicationScreen,
)


class NanopublicationFacet(Facet[Widget]):
    """Render a nanopublication."""

    def show(self) -> Widget:
        """Render a nanopublication."""
        return NanopublicationScreen(uri=self.this)



from __future__ import annotations

import re
from importlib import metadata
from typing import Type

from packageurl import PackageURL
from rdflib import URIRef

from iolanta import Facet
from iolanta.resolvers.base import Resolver

ENTRY_POINT_GROUP = 'iolanta.facets'


def _norm(name: str) -> str:
    """Normalize package name by replacing common separators with dashes."""
    return re.sub('[-_.]+', '-', name).lower()


class MissingFacetError(RuntimeError):
    """Raised when a required facet is not found in the installed packages."""


class PyPIResolver(Resolver):
    """Resolve facet IRIs into classes from PyPI packages."""

    def resolve(self, uri: URIRef) -> Type[Facet]:
        """Find and load a Facet class from a PyPI package."""
        dist_name, facet_name = self._get_package_info(uri)
        dist = self._get_distribution(dist_name, facet_name)
        entry_point = self._find_facet_entry_point(dist, facet_name)
        facet_class = entry_point.load()
        self._validate_facet_class(
            facet_class=facet_class,
            facet_name=facet_name,
            dist_name=dist.metadata['Name'],
        )
        return facet_class

    def _get_package_info(self, uri: URIRef) -> tuple[str, str]:
        """Extract and validate package information from URI."""
        package_url = PackageURL.from_string(str(uri))
        if package_url.type != 'pypi':
            raise NotImplementedError(f'{package_url.type} is not supported')

        dist_name = _norm(package_url.name or '')
        facet_name = package_url.subpath

        if not dist_name or not facet_name:
            raise ValueError('PURL must be pkg:pypi/<dist>#<FacetName>')

        return dist_name, facet_name

    def _get_distribution(
        self,
        dist_name: str,
        facet_name: str,
    ) -> metadata.Distribution:
        """Get package distribution and verify it exists."""
        try:
            return metadata.distribution(dist_name)
        except metadata.PackageNotFoundError as exc:
            raise MissingFacetError(
                f'This page requires `{dist_name}` (facet `{facet_name}`). '
                f'Install with: pip install "{dist_name}"',
            ) from exc

    def _find_facet_entry_point(
        self,
        dist: metadata.Distribution,
        facet_name: str,
    ) -> metadata.EntryPoint:
        """Find the facet entry point in the distribution."""
        matching_entry_points = [
            entry_point
            for entry_point in dist.entry_points
            if (
                entry_point.group == ENTRY_POINT_GROUP
                and entry_point.name == facet_name
            )
        ]

        if not matching_entry_points:
            dist_name = dist.metadata['Name']
            msg = (
                f'Facet `{facet_name}` not found in `{dist_name}` '
                f'{dist.version}. Ensure it exposes entry point '
                f'`{ENTRY_POINT_GROUP}` named `{facet_name}`.'
            )
            raise MissingFacetError(msg)

        return matching_entry_points[0]

    def _validate_facet_class(
        self,
        facet_class: Type[Facet],
        facet_name: str,
        dist_name: str,
    ) -> None:
        """Validate that the loaded class is a Facet subclass."""
        if not issubclass(facet_class, Facet):
            raise TypeError(
                f'Entry point `{facet_name}` in `{dist_name}` '
                'is not a subclass of iolanta.Facet',
            )



import functools
from dataclasses import dataclass, field, replace
from pathlib import Path
from typing import (  # noqa: WPS235
    Annotated,
    Any,
    Iterable,
    List,
    Mapping,
    Optional,
    Protocol,
    Set,
    Type,
)

import loguru
import yaml_ld
from pyparsing import ParseException
from rdflib import ConjunctiveGraph, Graph, Literal, URIRef
from rdflib.namespace import NamespaceManager
from rdflib.plugins.sparql.processor import SPARQLResult
from rdflib.term import Node
from yaml_ld.document_loaders.content_types import ParserNotFound
from yaml_ld.errors import YAMLLDError

from iolanta import entry_points, namespaces
from iolanta.conversions import path_to_iri
from iolanta.errors import UnresolvedIRI
from iolanta.facets.errors import FacetError
from iolanta.facets.facet import Facet
from iolanta.facets.locator import FacetFinder
from iolanta.models import ComputedQName, LDContext, NotLiteralNode
from iolanta.node_to_qname import node_to_qname
from iolanta.parse_quads import parse_quads
from iolanta.plugin import Plugin
from iolanta.query_result import (
    QueryResult,
    SPARQLParseException,
    SPARQLQueryArgument,
    format_query_bindings,
)
from iolanta.resolvers.base import Resolver
from iolanta.resolvers.dispatch import SchemeDispatchResolver
from iolanta.resolvers.pypi import PyPIResolver
from iolanta.resolvers.python_import import PythonImportResolver
from iolanta.sparqlspace.processor import normalize_term


class LoggerProtocol(Protocol):
    """Abstract Logger interface that unites `loguru` & standard `logging`."""

    def info(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log an INFO message."""

    def error(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log an ERROR message."""

    def warning(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log a WARNING message."""


def _create_default_graph():
    return ConjunctiveGraph(identifier=namespaces.LOCAL.term('_inference'))


@dataclass
class Iolanta:   # noqa: WPS214
    """Iolanta is a Semantic web browser."""

    language: Literal = Literal('en')

    project_root: Annotated[
        Path | None,
        (
            'File or directory the contents of which '
            'Iolanta will automatically load into the graph.'
        ),
    ] = None
    graph: ConjunctiveGraph = field(default_factory=_create_default_graph)
    force_plugins: List[Type[Plugin]] = field(default_factory=list)

    facet_resolver: Resolver = field(
        default_factory=functools.partial(
            SchemeDispatchResolver,
            python=PythonImportResolver,
            pkg=PyPIResolver,
        ),
    )

    logger: LoggerProtocol = loguru.logger

    could_not_retrieve_nodes: Set[Node] = field(
        default_factory=set,
        init=False,
    )

    @property
    def plugin_classes(self) -> List[Type[Plugin]]:
        """Installed Iolanta plugins."""
        return self.force_plugins or entry_points.plugins('iolanta.plugins')

    @functools.cached_property
    def plugins(self) -> List[Plugin]:
        """Construct a list of installed plugin instances."""
        return [
            plugin_class(iolanta=self)
            for plugin_class in self.plugin_classes
        ]

    def query(
        self,
        query_text: str,
        **kwargs: SPARQLQueryArgument,
    ) -> QueryResult:
        """Run a SPARQL `SELECT`, `CONSTRUCT`, or `ASK` query."""
        try:
            sparql_result: SPARQLResult = self.graph.query(
                query_text,
                processor='sparqlspace',
                initBindings=kwargs,
            )
        except SyntaxError as err:
            raise SPARQLParseException(
                error=err,
                query=query_text,
            ) from err

        if sparql_result.askAnswer is not None:
            return sparql_result.askAnswer

        if sparql_result.graph is not None:
            graph: Graph = sparql_result.graph
            for prefix, namespace in self.graph.namespaces():
                graph.bind(prefix, namespace)

            return graph

        return format_query_bindings(sparql_result.bindings)

    def reset(self):
        """Reset Iolanta graph."""
        self.graph = _create_default_graph()   # noqa: WPS601
        self.__post_init__()

    def add(  # noqa: C901, WPS231, WPS210, WPS213
        self,
        source: Path,
        context: Optional[LDContext] = None,
        graph_iri: Optional[URIRef] = None,
    ) -> 'Iolanta':
        """Parse & load information from given URL into the graph."""
        self.logger.info(f'Adding to graph: {source}')

        if not isinstance(source, Path):
            source = Path(source)

        for source_file in list(source.rglob('*')) or [source]:
            if source_file.is_dir():
                continue

            try:  # noqa: WPS225
                ld_rdf = yaml_ld.to_rdf(source_file)
            except ConnectionError as name_resolution_error:
                self.logger.warning(
                    '%s | name resolution error: %s',
                    source_file,
                    str(name_resolution_error),
                )
                continue
            except ParserNotFound as parser_not_found:
                self.logger.error(f'{source} | {parser_not_found}')
                continue
            except YAMLLDError as yaml_ld_error:
                # .add() only processes local files, so errors should be raised
                self.logger.error(f'{source_file} | {yaml_ld_error}')
                raise
            except ValueError as value_error:
                self.logger.error(f'{source} | {value_error}')
                continue

            self.logger.info(f'{source_file} is loaded.')

            graph = path_to_iri(source_file)
            try:
                quads = list(
                    parse_quads(
                        quads_document=ld_rdf,
                        graph=graph,
                        blank_node_prefix=str(source),
                    ),
                )
            except UnresolvedIRI as err:
                raise replace(
                    err,
                    context=None,
                    iri=graph,
                )

            if not quads:
                self.logger.info(f'{source_file} | No data found')
                continue

            self.graph.addN(quads)

        return self

    def infer(self, closure_class=None) -> 'Iolanta':
        """Apply inference."""
        return self

    def bind_namespaces(self):
        """Bind namespaces."""
        self.graph.namespace_manager = NamespaceManager(
            self.graph,
            bind_namespaces='none',
        )
        self.graph.bind(prefix='local', namespace=namespaces.LOCAL)
        self.graph.bind(prefix='iolanta', namespace=namespaces.IOLANTA)
        self.graph.bind(prefix='rdf', namespace=namespaces.RDF)
        self.graph.bind(prefix='rdfs', namespace=namespaces.RDFS)
        self.graph.bind(prefix='owl', namespace=namespaces.OWL)
        self.graph.bind(prefix='foaf', namespace=namespaces.FOAF)
        self.graph.bind(prefix='schema', namespace=namespaces.SDO)
        self.graph.bind(prefix='vann', namespace=namespaces.VANN)
        self.graph.bind(prefix='np', namespace=namespaces.NP)
        self.graph.bind(prefix='dcterms', namespace=namespaces.DCTERMS)
        self.graph.bind(prefix='rdfg', namespace=namespaces.RDFG)

    @functools.cached_property
    def context_paths(self) -> Iterable[Path]:
        """Compile list of context files."""
        directory = Path(__file__).parent / 'data'

        yield directory / 'context.yaml'

        for plugin in self.plugins:
            if path := plugin.context_path:
                yield path

    def add_files_from_plugins(self):
        """Load files from plugins."""
        for plugin in self.plugins:
            try:
                self.add(plugin.data_files)
            except Exception as error:
                self.logger.error(
                    f'Cannot load {plugin} plugin data files: {error}',
                )

    @property
    def facet_classes(self):
        """Get all registered facet classes."""
        return entry_points.plugins('iolanta.facets')

    def add_files_from_facets(self):
        """Add files from all registered facets to the graph."""
        for facet_class in self.facet_classes:
            try:
                self.add(facet_class.META)
            except AttributeError:
                pass

    def __post_init__(self):
        """Initialize after instance creation."""
        self.bind_namespaces()
        self.add_files_from_plugins()
        self.add_files_from_facets()
        if self.project_root:
            self.add(self.project_root)

    def render(
        self,
        node: Node,
        as_datatype: NotLiteralNode,
    ) -> Any:
        """Find an Iolanta facet for a node and render it."""
        node = normalize_term(node)

        if not as_datatype:
            raise ValueError(
                f'Please provide the datatype to render {node} as.',
            )

        if isinstance(as_datatype, list):
            raise NotImplementedError('Got a list for as_datatype :(')

        found = FacetFinder(
            iolanta=self,
            node=node,
            as_datatype=as_datatype,
        ).facet_and_output_datatype

        facet_class = self.facet_resolver.resolve(found['facet'])

        facet = facet_class(
            this=node,
            iolanta=self,
            as_datatype=found['output_datatype'],
        )

        try:
            return facet.show()

        except Exception as err:
            raise FacetError(
                node=node,
                facet_iri=found['facet'],
                error=err,
            ) from err

    def render_all(
        self,
        node: Node,
        as_datatype: NotLiteralNode,
    ) -> Iterable[Any]:
        """Find all possible Iolanta facets for a node and render them."""
        choices = list(
            FacetFinder(
                iolanta=self,
                node=node,
                as_datatype=as_datatype,
            ).choices(),
        )

        pairs = [
            (self.facet_resolver[row['facet']], row['output_datatype'])
            for row in choices
        ]

        facet_instances = [
            facet_class(
                this=node,
                iolanta=self,
                as_datatype=output_datatype,
            )
            for facet_class, output_datatype in pairs
        ]

        for facet in facet_instances:
            try:
                yield facet.show()
            except Exception as err:
                raise FacetError(
                    node=node,
                    facet_iri=None,
                    error=err,
                ) from err

    def node_as_qname(self, node: Node):
        """Render node as a QName if possible."""
        qname = node_to_qname(node, self.graph)
        return f'{qname.namespace_name}:{qname.term}' if isinstance(
            qname,
            ComputedQName,
        ) else node



from rich.markdown import Markdown
from textual.containers import VerticalScroll
from textual.widgets import Label


class Description(VerticalScroll):
    """Free form textual description."""

    DEFAULT_CSS = """
    Description {
      padding: 4;
      padding-top: 1;
      padding-bottom: 1;
      max-height: 100%;
    }
    """  # noqa: WPS115

    def __init__(self, renderable: str | Markdown):
        """
        Initialize a Description widget with a renderable content.

        Args:
            renderable: Content to display, either as a string or
                Markdown object
        """
        self.renderable = renderable
        super().__init__()

    def compose(self):
        """Build and return the widget's component structure."""
        yield Label(self.renderable)



import dataclasses
import datetime
import re
import time
from pathlib import Path
from threading import Lock
from types import MappingProxyType
from typing import Any, Iterable, Mapping

import diskcache
import funcy
import loguru
import platformdirs
import reasonable
import requests
import yaml_ld
from nanopub import NanopubClient
from rdflib import ConjunctiveGraph, Dataset, Namespace, URIRef, Variable
from rdflib.namespace import RDF as original_RDF
from rdflib.plugins.sparql.algebra import translateQuery
from rdflib.plugins.sparql.evaluate import evalQuery
from rdflib.plugins.sparql.parser import parseQuery
from rdflib.plugins.sparql.parserutils import CompValue
from rdflib.plugins.sparql.sparql import Query
from rdflib.query import Processor
from rdflib.term import BNode, Literal, Node
from requests.exceptions import ConnectionError, HTTPError, InvalidSchema
from yaml_ld.document_loaders.content_types import ParserNotFound
from yaml_ld.errors import NotFound, YAMLLDError
from yarl import URL

from iolanta.errors import UnresolvedIRI
from iolanta.namespaces import (  # noqa: WPS235
    DC,
    DCTERMS,
    FOAF,
    IOLANTA,
    LOCAL,
    META,
    OWL,
    PROV,
    RDF,
    RDFS,
    VANN,
)
from iolanta.parse_quads import NORMALIZE_TERMS_MAP, parse_quads

REASONING_ENABLED = True
OWL_REASONING_ENABLED = False

INFERENCE_DIR = Path(__file__).parent / 'inference'
INDICES = [
    URIRef('https://iolanta.tech/visualizations/index.yaml'),
]


REDIRECTS = MappingProxyType({
    # FIXME This is presently hardcoded; we need to
    #   - either find a way to resolve these URLs automatically,
    #   - or create a repository of those redirects online.
    'https://purl.org/vocab/vann/': URIRef(
        'https://vocab.org/vann/vann-vocab-20100607.rdf',
    ),
    URIRef(DC): URIRef(DCTERMS),
    URIRef(RDF): URIRef(RDF),
    URIRef(RDFS): URIRef(RDFS),
    URIRef(OWL): URIRef(OWL),

    # This one does not answer via HTTPS :(
    URIRef('https://xmlns.com/foaf/0.1/'): URIRef('http://xmlns.com/foaf/0.1/'),
    URIRef('https://www.nanopub.org/nschema'): URIRef(
        'https://www.nanopub.net/nschema#',
    ),
    URIRef('https://nanopub.org/nschema'): URIRef(
        'https://nanopub.net/nschema#',
    ),
    URIRef(PROV): URIRef('https://www.w3.org/ns/prov-o'),

    # Convert lexvo.org/id URLs to lexvo.org/data URLs
    r'https://lexvo\.org/id/(.+)': r'http://lexvo.org/data/\1',
    r'https://www\.lexinfo\.net/(.+)': r'http://www.lexinfo.net/\1',
})


@diskcache.Cache(
    directory=str(
        platformdirs.user_cache_path(
            appname='iolanta',
        ) / 'find_retractions_for',
    ),
).memoize(expire=datetime.timedelta(days=8).total_seconds())
def find_retractions_for(nanopublication: URIRef) -> set[URIRef]:
    """Find nanopublications that retract the given one."""
    # See https://github.com/fair-workflows/nanopub/issues/168 for
    # context of this dirty hack.
    use_server = 'http://grlc.nanopubs.lod.labs.vu.nl/api/local/local/'

    client = NanopubClient(use_server=use_server)
    client.grlc_urls = [use_server]

    http_url = str(nanopublication).replace(
        'https://',
        'http://',
    )

    try:
        retractions = client.find_retractions_of(http_url)
    except (requests.HTTPError, InvalidSchema):
        return set()

    return {URIRef(retraction) for retraction in retractions}


def _extract_from_mapping(  # noqa: WPS213
    algebra: Mapping[str, Any],
) -> Iterable[URIRef | Variable]:
    match algebra.name:
        case 'SelectQuery' | 'AskQuery' | 'Project' | 'Distinct' | 'Slice':
            yield from extract_mentioned_urls(algebra['p'])

        case 'BGP':
            yield from [   # noqa: WPS353, WPS221
                term
                for triple in algebra['triples']
                for term in triple
                if not isinstance(term, Literal)
            ]

        case 'Filter' | 'UnaryNot' | 'OrderCondition':
            yield from extract_mentioned_urls(algebra['expr'])   # noqa: WPS204

        case built_in if built_in.startswith('Builtin_'):
            yield from extract_mentioned_urls(algebra['arg'])

        case 'RelationalExpression':
            yield from extract_mentioned_urls(algebra['expr'])
            yield from extract_mentioned_urls(algebra['other'])

        case 'LeftJoin':
            yield from extract_mentioned_urls(algebra['p1'])
            yield from extract_mentioned_urls(algebra['p2'])
            yield from extract_mentioned_urls(algebra['expr'])

        case 'Join':
            yield from extract_mentioned_urls(algebra['p1'])
            yield from extract_mentioned_urls(algebra['p2'])

        case 'ConditionalOrExpression' | 'ConditionalAndExpression':
            yield from extract_mentioned_urls(algebra['expr'])
            yield from extract_mentioned_urls(algebra['other'])

        case 'OrderBy':
            yield from extract_mentioned_urls(algebra['p'])
            yield from extract_mentioned_urls(algebra['expr'])

        case 'TrueFilter':
            return

        case 'Graph':
            yield from extract_mentioned_urls(algebra['p'])
            yield from extract_mentioned_urls(algebra['term'])

        case unknown_name:
            formatted_keys = ', '.join(algebra.keys())
            loguru.logger.info(
                'Unknown SPARQL expression '
                f'{unknown_name}({formatted_keys}): {algebra}',
            )
            return


def extract_mentioned_urls(
    algebra,
) -> Iterable[URIRef | Variable]:
    """Extract flat triples from parsed SPARQL query."""
    match algebra:
        case Variable() as query_variable:
            yield query_variable

        case URIRef() as uri_ref:
            yield uri_ref

        case dict():
            yield from _extract_from_mapping(algebra)

        case list() as expressions:
            for expression in expressions:
                yield from extract_mentioned_urls(expression)

        case unknown_algebra:
            algebra_type = type(unknown_algebra)
            raise ValueError(
                f'Algebra of unknown type {algebra_type}: {unknown_algebra}',
            )


def normalize_term(term: Node) -> Node:
    """
    Normalize RDF terms.

    This is an exctremely dirty hack to fix a bug in OWL reported here:

    > https://stackoverflow.com/q/78934864/1245471

    TODO This is:
      * A dirty hack;
      * Based on hard code.
    """
    return NORMALIZE_TERMS_MAP.get(term, term)


def resolve_variables(
    terms: Iterable[URIRef | Variable],
    bindings: Mapping[str, Node],
):
    """Replace variables with their values."""
    for term in terms:
        match term:
            case URIRef():
                yield term

            case Variable() as query_variable:
                variable_value = bindings.get(str(query_variable))
                if (
                    variable_value is not None
                    and isinstance(variable_value, URIRef)
                ):
                    yield variable_value


def extract_mentioned_urls_from_query(
    query: Query,
    bindings: dict[str, Node],
    base: str | None,
    namespaces: dict[str, Namespace],
) -> tuple[Query, set[URIRef]]:
    """Extract URLs that a SPARQL query somehow mentions."""
    return query, set(
        resolve_variables(
            extract_mentioned_urls(query.algebra),
            bindings=bindings,
        ),
    )


@dataclasses.dataclass
class Loaded:
    """The data was loaded successfully."""


@dataclasses.dataclass
class Skipped:
    """The data is already in the graph and loading was skipped."""


LoadResult = Loaded | Skipped


def _extract_nanopublication_uris(
    algebra: CompValue,
) -> Iterable[URIRef]:
    """Extract nanopublications to get retracting information for."""
    match algebra.name:
        case 'SelectQuery' | 'AskQuery' | 'Project' | 'Distinct' | 'Graph':
            yield from _extract_nanopublication_uris(algebra['p'])
        case 'ConstructQuery':
            # CONSTRUCT queries don't have nanopublication URIs in bindings
            return

        case 'Slice':
            yield from _extract_nanopublication_uris(algebra['p'])

        case 'BGP':
            for retractor, retracts, retractee in algebra['triples']:
                if retracts == URIRef(
                    'https://purl.org/nanopub/x/retracts',
                ) and isinstance(retractor, Variable):
                    yield retractee

        case 'LeftJoin' | 'Join':
            yield from _extract_nanopublication_uris(algebra['p1'])
            yield from _extract_nanopublication_uris(algebra['p2'])

        case 'Filter' | 'OrderBy':
            return

        case unknown_name:
            raise ValueError(
                f'Unknown algebra name: {unknown_name}, '
                f'content: {algebra}',
            )


def apply_redirect(source: URIRef) -> URIRef:   # noqa: WPS210
    """
    Rewrite the URL using regex patterns and group substitutions.

    For each pattern in REDIRECTS:
    - If the pattern matches the source URI
    - Replace the source with the destination, substituting any regex groups
    """
    source_str = str(source)

    for pattern, destination in REDIRECTS.items():
        pattern_str = str(pattern)
        destination_str = str(destination)

        match = re.match(pattern_str, source_str)
        if match:
            # Replace any group references in the destination
            # (like \1, \2, etc.)
            redirected_uri = re.sub(
                pattern_str,
                destination_str,
                source_str,
            )
            return URIRef(redirected_uri)

    return source


def extract_triples(algebra: CompValue) -> Iterable[tuple[Node, Node, Node]]:
    """Extract triples from a SPARQL query algebra instance."""
    if isinstance(algebra, CompValue):
        for key, value in algebra.items():  # noqa: WPS110
            if key == 'triples':
                yield from value

            else:
                yield from extract_triples(value)


@dataclasses.dataclass(frozen=True)
class NanopubQueryPlugin:
    """Import additional information from Nanopublications Registry."""

    graph: Dataset

    def __call__(self, query: Query, bindings: dict[str, Any]):
        """Get stuff from Nanopub Registry, if it makes sense."""
        class_uris = resolve_variables(
            set(self._find_classes(query.algebra)),
            bindings,
        )
        for class_uri in class_uris:
            if self._is_from_nanopubs(class_uri):
                self._load_instances(class_uri)

    def _find_classes(self, algebra: CompValue) -> Iterable[URIRef]:
        triples = extract_triples(algebra)
        for _subject, potential_type, potential_class in triples:
            if potential_type == original_RDF.type:
                yield potential_class

    @funcy.retry(errors=requests.HTTPError, tries=3, timeout=3)
    def _load_instances(self, class_uri: URIRef):
        """
        Load instances from Nanopub Registry.

        FIXME: Can we cache this?
        """
        response = requests.post(   # noqa: S113
            'https://query.knowledgepixels.com/repo/full',
            data={
                'query': 'CONSTRUCT WHERE { ?instance a <%s> }' % class_uri,
            },
            headers={
                'Accept': 'application/ld+json',
            },
        )

        response.raise_for_status()

        self.graph.get_context(BNode()).parse(
            data=response.text,
            format='json-ld',
        )

    def _is_from_nanopubs(self, class_uri: URIRef) -> bool:
        if not isinstance(class_uri, URIRef):
            raise ValueError(f'Not a URIRef: {class_uri}')

        return self.graph.query(   # noqa: WPS462
            """
            ASK WHERE {
                ?_whatever <https://purl.org/nanopub/x/introduces> $class
            }
            """,
            initBindings={
                'class': class_uri,
            },
        ).askAnswer


@dataclasses.dataclass(frozen=True)
class GlobalSPARQLProcessor(Processor):  # noqa: WPS338, WPS214
    """
    Execute SPARQL queries against the whole Linked Data Web, or The Cyberspace.

    When running the queries, we will try to find and to import pieces of LD
    which can be relevant to the query we are executing.
    """

    graph: ConjunctiveGraph
    inference_lock: Lock = dataclasses.field(default_factory=Lock)
    logger: Any = loguru.logger

    def __post_init__(self):
        """Note that we do not presently need OWL inference."""
        self.graph.last_not_inferred_source = None
        self.graph._indices_loaded = False

    def _maybe_load_indices(self):
        if not self.graph._indices_loaded:
            for index in INDICES:
                self.load(index)

            self.graph._indices_loaded = True

    def query(   # noqa: WPS211, WPS210, WPS231, C901
        self,
        strOrQuery,
        initBindings=None,
        initNs=None,
        base=None,
        DEBUG=False,
    ):
        """
        Evaluate a query with the given initial bindings, and initial
        namespaces. The given base is used to resolve relative URIs in
        the query and will be overridden by any BASE given in the query.
        """
        self._maybe_load_indices()

        initBindings = initBindings or {}
        initNs = initNs or {}
        
        if isinstance(strOrQuery, Query):
            query = strOrQuery

        else:
            parse_tree = parseQuery(strOrQuery)
            query = translateQuery(parse_tree, base, initNs)

        # Only extract nanopublications from SELECT/ASK queries, not CONSTRUCT
        if query.algebra.name != 'ConstructQuery':
            self.load_retracting_nanopublications_by_query(
                query=query,
                bindings=initBindings,
                base=base,
                namespaces=initNs,
            )

        query, urls = extract_mentioned_urls_from_query(
            query=query,
            bindings=initBindings,
            base=base,
            namespaces=initNs,
        )

        # Filter out inference graph names (they're not URLs to load)
        urls = {url for url in urls if not str(url).startswith('inference:')}

        for url in urls:
            try:
                self.load(url)
            except Exception as err:
                self.logger.exception(f'Failed to load {url}: {err}', url, err)

        # Run inference if there's new data since last inference run
        # (after URLs are loaded so inference can use the loaded data)
        if self.graph.last_not_inferred_source is not None:
            self.logger.debug(f'Running inference, last_not_inferred_source: {self.graph.last_not_inferred_source}')
            self._run_inference()
        else:
            self.logger.debug('Skipping inference, last_not_inferred_source is None')

        NanopubQueryPlugin(graph=self.graph)(query, bindings=initBindings)

        is_anything_loaded = True
        while is_anything_loaded:
            is_anything_loaded = False

            query_result = evalQuery(self.graph, query, initBindings, base)

            try:
                bindings = list(query_result['bindings'])
            except KeyError:
                # This was probably an ASK query
                return query_result

            for row in bindings:
                break
                for _, maybe_iri in row.items():
                    if (
                        isinstance(maybe_iri, URIRef)
                        and isinstance(self.load(maybe_iri), Loaded)
                    ):
                        is_anything_loaded = True   # noqa: WPS220
                        self.logger.info(   # noqa: WPS220
                            'Newly loaded: {uri}',
                            uri=maybe_iri,
                        )

        query_result['bindings'] = bindings
        return query_result

    def _is_loaded(self, uri: URIRef) -> bool:
        """Find out if this URI in the graph already."""
        return funcy.first(
            self.graph.quads((
                uri,
                IOLANTA['last-loaded-time'],
                None,
                META,
            )),
        ) is not None

    def _mark_as_loaded(self, uri: URIRef):
        self.graph.add((
            uri,
            IOLANTA['last-loaded-time'],
            Literal(datetime.datetime.now()),
            META,
        ))

    def _follow_is_visualized_with_links(self, uri: URIRef):
        """Follow `dcterms:isReferencedBy` links."""
        triples = self.graph.triples((uri, DCTERMS.isReferencedBy, None))
        for _, _, visualization in triples:
            if isinstance(visualization, URIRef):
                self.load(visualization)

    def load(   # noqa: C901, WPS210, WPS212, WPS213, WPS231
        self,
        source: URIRef,
    ) -> LoadResult:
        """
        Try to load LD denoted by the given `source`.

        TODO This function is too big, we have to refactor it.
        """
        url = URL(source)

        if url.scheme in {'file', 'python', 'local', 'urn', 'doi'}:
            # FIXME temporary fix. `yaml-ld` doesn't read `context.*` files and
            #   fails.
            return Skipped()

        if url.fragment:
            # Fragment on an HTML page resolves to that same page. Let us remove
            # this ambiguity, then.
            # TODO: It works differently for JSON-LD documents AFAIK. Need to
            #   double check that.
            url = url.with_fragment(None)
            source = URIRef(str(f'{url}#'))

        self._follow_is_visualized_with_links(source)

        new_source = apply_redirect(source)
        if new_source != source:
            self.logger.info(
                'Rewriting: {source} → {new_source}',
                source=source,
                new_source=new_source,
            )
            return self.load(new_source)

        source_uri = normalize_term(source)
        if self._is_loaded(source_uri):
            return Skipped()

        # FIXME This is definitely inefficient. However, python-yaml-ld caches
        #   the document, so the performance overhead is not super high.
        try:
            _resolved_source = yaml_ld.load_document(source)['documentUrl']
        except NotFound as not_found:
            self.logger.info(f'{not_found.path} | 404 Not Found')
            namespaces = [RDF, RDFS, OWL, FOAF, DC, VANN]

            for namespace in namespaces:
                if not_found.path.startswith(str(namespace)):
                    self.load(URIRef(namespace))
                    self.logger.info(
                        'Redirecting %s → namespace %s',
                        not_found.path,
                        namespace,
                    )
                    return Loaded()

            self.logger.info(
                '{path} | Cannot find a matching namespace',
                path=not_found.path,
            )

            self.graph.add((
                source_uri,
                RDF.type,
                IOLANTA['not-found'],
                source_uri,
            ))

            self._mark_as_loaded(source_uri)

            return Loaded()

        except Exception as err:
            self.logger.info(f'{source} | Failed: {err}')
            self.graph.add((
                URIRef(source),
                RDF.type,
                IOLANTA['failed'],
                source_uri,
            ))
            self._mark_as_loaded(source_uri)

            return Loaded()

        if _resolved_source:
            _resolved_source_uri_ref = URIRef(_resolved_source)
            if _resolved_source_uri_ref != URIRef(source):
                self.graph.add((
                    source_uri,
                    IOLANTA['redirects-to'],
                    _resolved_source_uri_ref,
                ))
                source = _resolved_source

        self._mark_as_loaded(source_uri)

        try:  # noqa: WPS225
            ld_rdf = yaml_ld.to_rdf(source)
        except ConnectionError as name_resolution_error:
            self.logger.info(
                '%s | name resolution error: %s',
                source,
                str(name_resolution_error),
            )
            return Loaded()
        except ParserNotFound as parser_not_found:
            self.logger.info(f'{source} | {parser_not_found}')
            return Loaded()
        except YAMLLDError as yaml_ld_error:
            self.logger.error(f'{source} | {yaml_ld_error}')
            return Loaded()
        except HTTPError as http_error:
            self.logger.warning(f'{source} | HTTP error: {http_error}')
            return Loaded()

        try:
            quads = list(
                parse_quads(
                    quads_document=ld_rdf,
                    graph=source_uri,
                    blank_node_prefix=str(source),
                ),
            )
        except UnresolvedIRI as err:
            raise dataclasses.replace(
                err,
                context=None,
                iri=source,
            )

        if not quads:
            self.logger.info('{source} | No data found', source=source)
            return Loaded()

        self.graph.addN(quads)
        self.graph.last_not_inferred_source = source

        into_graphs = ', '.join({
            quad.graph
            for quad in quads
        })
        self.logger.info(
            f'{source} | loaded {len(quads)} triples into graphs: {into_graphs}',
        )

        return Loaded()

    def resolve_term(self, term: Node, bindings: dict[str, Node]):
        """Resolve triple elements against initial variable bindings."""
        if isinstance(term, Variable):
            return bindings.get(
                str(term),
                term,
            )

        return term

    def _run_inference(self):  # noqa: WPS231
        """
        Run inference queries from the inference directory.

        For each SPARQL file in the inference directory:
        1. Truncate the named graph `local:inference-{filename}`
        2. Execute the CONSTRUCT query
        3. Insert the resulting triples into that graph
        """
        with self.inference_lock:
            for inference_file in INFERENCE_DIR.glob('*.sparql'):
                filename = inference_file.stem  # filename without .sparql extension
                inference_graph = URIRef(f'inference:{filename}')

                # Truncate the inference graph
                context = self.graph.get_context(inference_graph)
                context.remove((None, None, None))

                # Read and execute the CONSTRUCT query
                query_text = inference_file.read_text()
                result = self.graph.query(query_text)

                # CONSTRUCT queries return a SPARQLResult with a graph attribute
                result_graph = result.get('graph') if isinstance(result, dict) else result.graph
                self.logger.debug(f'Inference {filename}: result_graph is {result_graph}, type: {type(result_graph)}')
                if result_graph is not None:
                    inferred_quads = [
                        (s, p, o, inference_graph)
                        for s, p, o in result_graph
                    ]
                    self.logger.debug(f'Inference {filename}: generated {len(inferred_quads)} quads')

                    if inferred_quads:
                        self.graph.addN(inferred_quads)
                        self.logger.info(
                            'Inference {filename}: added {count} triples',
                            filename=filename,
                            count=len(inferred_quads),
                        )
                else:
                    self.logger.debug(f'Inference {filename}: result_graph is None')
        
        # Clear the flag after running inference
        self.graph.last_not_inferred_source = None

    def load_retracting_nanopublications_by_query(    # noqa: WPS231
        self,
        query: Query,
        bindings: dict[str, Node],
        base: str | None,
        namespaces: dict[str, Namespace],
    ):
        """
        If the query requires information about retracting nanopubs, load them.

        FIXME: This function presently does nothing because `nanopub` library
          has problems: https://github.com/fair-workflows/nanopub/issues/168

        TODO: Generalize this mechanism to allow for plugins which analyze
          SPARQL queries and load information into the graph based on their
          content.
        """
        nanopublications = list(
            resolve_variables(
                terms=_extract_nanopublication_uris(query.algebra),
                bindings=bindings,
            ),
        )

        for nanopublication in nanopublications:
            retractions = find_retractions_for(nanopublication)
            for retraction in retractions:
                self.load(retraction)



import funcy
from rdflib import Literal
from textual.containers import VerticalScroll

from iolanta.facets.page_title import PageTitle
from iolanta.facets.textual_nanopublication.term_list_widget import TermList
from iolanta.facets.textual_nanopublication.term_widget import TermWidget
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES, DCTERMS, NP
from iolanta.widgets.mixin import IolantaWidgetMixin

NANOPUBLICATION_QUERY = """
SELECT ?assertion ?author ?created_time ?retract WHERE {
    $uri np:hasAssertion ?assertion .

    OPTIONAL {
        $uri (
            dcterms:creator
            | <https://purl.org/pav/createdBy>
            | <https://swan.mindinformatics.org/ontologies/1.2/pav/createdBy>
        ) ?author .
    }

    OPTIONAL {
        $uri dcterms:created ?created_time .
    }

    OPTIONAL {
        GRAPH ?retracting_assertion {
            ?retractor <https://purl.org/nanopub/x/retracts> $uri .
        }

        ?retract np:hasAssertion ?retracting_assertion .
    }
}
"""


class NanopublicationScreen(IolantaWidgetMixin, VerticalScroll):
    """Nanopublication screen."""

    def __init__(self, uri: NotLiteralNode):
        """Initialize."""
        self.uri = uri
        super().__init__()

    def compose(self):
        """Render components of the nanopublication screen."""
        yield PageTitle(NP.Nanopublication)

        row = funcy.first(
            self.iolanta.query(   # noqa: WPS462
                NANOPUBLICATION_QUERY,
                uri=self.uri,
            ),
        )

        if not row:
            return

        yield self.iolanta.render(
            row['assertion'],
            as_datatype=DATATYPES['textual-graph-triples'],
        )

        provenance = []
        if row.get('author'):
            provenance.extend([
                TermWidget(DCTERMS.creator, as_datatype=DATATYPES.icon),
                TermWidget(row['author']),
            ])

        if row.get('created_time'):
            provenance.append(
                TermWidget(row['created_time']),
            )

        if retract := row.get('retract'):
            provenance.extend([
                TermWidget(Literal('Retracted by'), background_color='darkred'),
                TermWidget(retract, background_color='darkred'),
            ])

        if provenance:
            yield TermList(provenance)



import re
from enum import Enum
from typing import Any, Dict, List, NamedTuple, Union

from rdflib import Literal, URIRef
from rdflib.term import BNode, Node, Variable


class ComputedQName(NamedTuple):
    """QName computed from an URIRef."""

    namespace_name: str
    namespace_url: str
    term: str


class QueryResultsFormat(str, Enum):
    """Format to print query results in CLI."""

    PRETTY = 'pretty'
    CSV = 'csv'
    JSON = 'json'


# JSON-LD context
LDContext = Union[Dict[str, Any], List[Dict[str, Any]]]   # type: ignore
LDDocument = Union[Dict[str, Any], List[Dict[str, Any]]]  # type: ignore

NotLiteralNode = Union[URIRef, BNode]

# Named context URLs
ContextAliases = Dict[str, str]


def render_node(node: Node) -> str:
    """Render an RDFLib node as string."""
    if isinstance(node, URIRef):
        return f'<{node}>'

    if isinstance(node, Literal):
        rendered_node = f'"{node}"'
        if node.datatype:
            rendered_node = f'{rendered_node}^^{node.datatype}'

        if node.language:
            rendered_node = f'{rendered_node}@{node.language}'

        return rendered_node

    return f'<??? What is this? {node} ???>'


class Triple(NamedTuple):
    """RDF triple."""

    subject: NotLiteralNode
    predicate: NotLiteralNode
    object: URIRef | Literal  # noqa: WPS125

    def as_quad(self, graph: URIRef) -> 'Quad':
        """Add graph to this triple and hence get a quad."""
        return Quad(
            subject=self.subject,
            predicate=self.predicate,
            object=self.object,
            graph=graph,
        )


class TripleWithVariables(NamedTuple):
    """RDF triple."""

    subject: NotLiteralNode | Variable
    predicate: NotLiteralNode | Variable
    object: Node | Variable



class TripleTemplate(NamedTuple):
    subject: NotLiteralNode | None
    predicate: NotLiteralNode | None
    object: Node | None


def _normalize_term(term: Node):
    return term


class Quad(NamedTuple):
    """Triple assigned to a named graph."""

    subject: Node
    predicate: Node
    object: Node  # noqa: WPS125
    graph: URIRef

    def as_triple(self):
        """Convert this to triple."""
        return Triple(self.subject, self.predicate, self.object)

    def __repr__(self):
        """Represent the quad as string."""
        rendered_subject = render_node(self.subject)
        rendered_predicate = render_node(self.predicate)
        rendered_object = render_node(self.object)
        rendered_graph = render_node(self.graph)

        return (
            f'({rendered_subject} {rendered_predicate} {rendered_object} @ '
            f'{rendered_graph})'  # noqa: WPS326
        )

    def replace(self, mapping: dict[Node, URIRef]):
        """Replace variables in the quad."""
        terms = [
            mapping.get(term, term)
            for term in self
        ]

        return Quad(*terms)

    def normalize(self) -> 'Quad':
        """Normalize the quad by applying normalization to all its terms."""
        terms = [
            _normalize_term(term)
            for term in self
        ]
        return Quad(*terms)



from rich.style import Style
from rich.text import Text

from iolanta.facets.cli import Renderable, RichFacet
from iolanta.facets.generic.default import DefaultMixin


class Default(DefaultMixin, RichFacet):
    def render_link(self) -> Renderable:
        return Text(
            self.render_label(),
            style=Style(link=self.description.url),
        )



import logging
from concurrent.futures import ThreadPoolExecutor

from rdflib.term import Node
from rich.console import RenderableType
from textual.app import App, ComposeResult
from textual.css.query import NoMatches
from textual.widgets import Footer, Header

from iolanta.facets.textual_browser.page_switcher import (
    ConsoleSwitcher,
    DevConsole,
    PageSwitcher,
)
from iolanta.iolanta import Iolanta

POPUP_TIMEOUT = 30   # seconds


class DevConsoleHandler(logging.Handler):
    """Pipe log output → dev console."""

    def __init__(self, console: DevConsole, level=logging.NOTSET) -> None:
        """Set parameters."""
        self.console = console
        super().__init__(level=level)

    def emit(self, record: logging.LogRecord) -> None:
        """Write a message when invoked by `logging`."""
        message = self.format(record)
        self.console.write(message)


def _log_message_to_dev_console(app: App):
    """Log a message to the dev console."""
    def log_message_to_dev_console(message: str):   # noqa: WPS430
        try:
            app.query_one(DevConsole).write(message)
        except NoMatches:
            return

    return log_message_to_dev_console


class IolantaBrowser(App):  # noqa: WPS214, WPS230
    """Browse Linked Data."""

    def __init__(self, iolanta: Iolanta, this: Node):
        """Set up parameters for the browser."""
        self.iolanta = iolanta
        self.this = this
        self.renderers = ThreadPoolExecutor()
        super().__init__()

    BINDINGS = [  # noqa: WPS115
        ('t', 'toggle_dark', 'Toggle Dark Mode'),
        ('q', 'quit', 'Quit'),
    ]

    def compose(self) -> ComposeResult:
        """Compose widgets."""
        yield Header(icon='👁️')
        yield Footer()
        yield ConsoleSwitcher()

    def on_mount(self):
        """Set title."""
        self.title = 'Iolanta'

        logging.basicConfig(
            level=logging.INFO,
            handlers=[
                DevConsoleHandler(
                    console=self.query_one(DevConsole),
                    level=logging.INFO,
                ),
            ],
            force=True,
        )

        # Disable stderr logging, to not break the TUI.
        # Remove only the stderr handler, keep file handler
        for handler_id in list(self.iolanta.logger._core.handlers.keys()):
            handler = self.iolanta.logger._core.handlers[handler_id]
            if hasattr(handler, 'sink') and str(handler.sink) == '<stderr>':
                self.iolanta.logger.remove(handler_id)

        # Log to the dev console.
        self.iolanta.logger.add(
            _log_message_to_dev_console(self),
            level='INFO',
            format='{time} {level} {message}',
        )

        self.iolanta.logger.add(
            lambda msg: self.notify(
                msg,
                severity='warning',
                timeout=POPUP_TIMEOUT,
            ),
            level='WARNING',
            format='{message}',
        )

    def action_toggle_dark(self) -> None:
        """Toggle dark mode."""
        self.dark = not self.dark

    def action_goto(
        self,
        destination: str,
        facet_iri: str | None = None,
    ):
        """Go to an IRI."""
        self.query_one(PageSwitcher).action_goto(destination, facet_iri)

    def dev_console_log(self, renderable: RenderableType | object):
        """Print a renderable to the dev console."""
        self.query_one(DevConsole).write(renderable)



from typing import Iterable

import funcy
from rdflib.term import Literal, Node
from textual.containers import Horizontal, Vertical
from textual.widget import Widget

from iolanta import Facet
from iolanta.facets.textual_nanopublication.term_widget import TermWidget
from iolanta.models import Triple
from iolanta.widgets.mixin import IolantaWidgetMixin

BACKGROUND_COLORS = [    # noqa: WPS407
    # Let's honor 🇦🇲 flag!
    'darkred',
    'darkblue',
    'darkorange',

    'darkcyan',
    'darkgoldenrod',
    # 'darkgray',
    'darkgreen',
    # 'darkgrey',
    'darkkhaki',
    'darkmagenta',
    'darkolivegreen',
    'darkorchid',
    'darksalmon',
    'darkseagreen',
    # 'darkslateblue',
    # 'darkslategray',
    # 'darkslategrey',
    'darkturquoise',
    'darkviolet',
]


class TripleView(IolantaWidgetMixin, Horizontal):
    """Display a triple."""

    DEFAULT_CSS = """
    TripleView {
        padding-left: 0;
        padding-top: 0;
        padding-bottom: 1;
        height: auto;
    }
    """

    def __init__(
        self,
        triple: Triple,
        color_per_node: dict[Node, str] | None = None,
    ):
        """Initialize."""
        self.triple = triple
        self.color_per_node = color_per_node
        super().__init__()

    def compose(self):
        """Render the triple."""
        for term in self.triple:   # noqa: WPS526
            yield TermWidget(
                term,
                background_color=self.color_per_node.get(term),
            )


def construct_color_per_node(nodes: Iterable[Node]) -> dict[Node, str]:
    """Distribute rainbow colors over the nodes and choose a color for each."""
    non_literal_nodes = filter(
        lambda node: not isinstance(node, Literal),
        nodes,
    )

    distinct_nodes = funcy.ldistinct(non_literal_nodes)

    return funcy.zipdict(
        distinct_nodes,
        funcy.cycle(BACKGROUND_COLORS),
    )


class TriplesView(IolantaWidgetMixin, Vertical):
    """Display a set of triples."""

    DEFAULT_CSS = """
    TriplesView {
        padding: 1 2;
        height: auto;
    }
    """

    def __init__(self, triples: Iterable[Triple]):
        """Initialize."""
        self.triples = triples
        self.color_per_node = construct_color_per_node(
            [
                term
                for triple in triples
                for term in triple
            ],
        )
        super().__init__()

    def compose(self):
        """Mount the triple stubs."""
        for triple in self.triples:   # noqa: WPS526
            yield TripleView(triple, color_per_node=self.color_per_node)

    def on_mount(self):
        """Initialize triples rendering."""
        self.run_worker(self.render_triples, thread=True)

    def render_triples(self):
        """Render triples."""
        term_view: TermWidget
        triple_view: TripleView
        for triple_view in self.children:
            for term_view in triple_view.children:
                term_view.renderable = self.iolanta.render(
                    term_view.uri,
                    as_datatype=term_view.as_datatype,
                )


class GraphTriplesFacet(Facet[Widget]):
    """Render a graph as triples."""

    def show(self) -> Widget:
        """Render a graph as triples."""
        rows = self.query(  # noqa: WPS462
            """
            SELECT ?subject ?predicate ?object WHERE {
                GRAPH $graph {
                    ?subject ?predicate ?object .
                }
            }
            ORDER BY ?subject ?predicate ?object
            """,
            graph=self.this,
        )

        triples = [
            Triple(
                subject=row['subject'],
                predicate=row['predicate'],
                object=row['object'],
            )
            for row in rows
        ]

        return TriplesView(triples)



from typing import ClassVar

from rdflib.term import Node
from rich.text import Text
from textual.app import RenderResult
from textual.binding import Binding, BindingType
from textual.events import Click
from textual.reactive import reactive
from textual.widget import Widget
from textual.worker import Worker, WorkerState

from iolanta.facets.textual_default.triple_uri_ref import TripleURIRef
from iolanta.facets.textual_nanopublication.models import TermPositionInTriple
from iolanta.models import NotLiteralNode, Triple
from iolanta.namespaces import DATATYPES
from iolanta.widgets.mixin import IolantaWidgetMixin


class TermWidget(   # noqa: WPS214
    IolantaWidgetMixin,
    Widget,
    can_focus=True,
    inherit_bindings=False,
):
    """Widget to display a term."""

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding('enter', 'goto', 'Goto'),
    ]

    DEFAULT_CSS = """
    TermWidget {
        width: auto;
        height: auto;
        padding: 1 2;
    }

    TermWidget:hover {
        background: $boost;
    }

    TermWidget:focus {
        background: darkslateblue;
    }
    """

    renderable: str | None = reactive[str | None](  # noqa: WPS465
        None,
        init=False,
        layout=True,
    )

    def __init__(
        self,
        uri: Node,
        as_datatype: NotLiteralNode = DATATYPES.title,
        triple: Triple | None = None,
        position: TermPositionInTriple | None = None,
        background_color: str | None = None,
    ):
        """Initialize parameters for rendering, navigation, & provenance."""
        self.uri = uri
        self.as_datatype = as_datatype
        self.triple = triple
        self.position_in_triple = position
        self.background_color = background_color
        super().__init__()
        qname = self.app.iolanta.node_as_qname(  # noqa: WPS601
            uri,
        )
        self.renderable = Text(  # noqa: WPS601
            f'⏳ {qname}',
            style='#696969',
        )

        if self.background_color:
            self.styles.background = self.background_color

    def on_worker_state_changed(self, event: Worker.StateChanged):
        """Show the title after it has been rendered."""
        match event.state:
            case WorkerState.SUCCESS:
                self.renderable = event.worker.result  # noqa: WPS601
                self.styles.color = 'white'

            case WorkerState.ERROR:
                raise ValueError(event)

    def render(self) -> RenderResult:
        """Render title of the node."""
        return self.renderable

    def action_goto(self):
        """Navigate."""
        self.app.action_goto(self.uri)

    def action_provenance(self):
        """Navigate to provenance for the property value."""
        self.app.action_goto(
            TripleURIRef.from_triple(
                subject=self.triple.subject,
                predicate=self.triple.predicate,
                object_=self.triple.object,
            ),
        )

    def on_click(self, event: Click):
        """
        Navigate to the property if we are focused.

        FIXME: Does not work; causes navigation even if not focused.
        """
        if self.has_focus:
            return self.action_goto()

    def on_focus(self):
        """Handle focus state."""
        if self.background_color:
            self.styles.background = 'darkslateblue'

    def on_blur(self):
        """Handle blur state."""
        if self.background_color:
            self.styles.background = self.background_color



import functools
from xml.dom import minidom  # noqa: S408

import funcy
from rdflib.term import Literal, Node
from rich.syntax import Syntax
from textual.containers import VerticalScroll
from textual.widget import Widget
from textual.widgets import Static

from iolanta.facets.facet import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.facets.textual_default.widgets import (
    LiteralPropertyValue,
    PropertiesContainer,
    PropertyName,
    PropertyRow,
    PropertyValue,
    PropertyValues,
)
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DC, RDFS, SDO


class TextualDefaultFacet(Facet[Widget]):   # noqa: WPS214
    """Default rendering engine."""

    query_file_name = 'properties.sparql'

    @functools.cached_property
    def grouped_properties(self) -> dict[NotLiteralNode, list[Node]]:
        """Properties of current node & their values."""
        property_rows = self.stored_query(
            self.query_file_name,
            iri=self.this,
        )

        property_pairs = [
            (row['property'], row['object'])
            for row in property_rows
        ]

        property_pairs = [
            (property_iri, object_node)
            for property_iri, object_node in property_pairs
            if (
                not isinstance(object_node, Literal)
                or not (language := object_node.language)  # noqa: W503
                or (language == self.iolanta.language)     # noqa: W503
            )
        ]

        return funcy.group_values(property_pairs)

    @functools.cached_property
    def rows(self):
        """Generate rows for the properties table."""
        for property_iri, property_values in self.grouped_properties.items():
            property_name = PropertyName(
                iri=property_iri,
                qname=self.iolanta.node_as_qname(property_iri),
            )

            property_values = [
                LiteralPropertyValue(
                    property_value=property_value,
                    subject=self.this,
                    property_iri=property_iri,
                ) if isinstance(property_value, Literal) else PropertyValue(
                    property_value=property_value,
                    subject=self.this,
                    property_iri=property_iri,
                    property_qname=self.iolanta.node_as_qname(property_iri),
                )
                for property_value in property_values
            ]

            yield PropertyRow(
                property_name,
                PropertyValues(*property_values),
            )

    @functools.cached_property
    def description(self) -> str | Syntax | None:
        """
        Candidates for description.

        FIXME: We mutate `grouped_properties` here.

        TODO: Move into a separate Facet.
        """
        choices = [
            description
            for description_property in [
                DC.description,
                SDO.description,
                RDFS.comment,
            ]
            for description in self.grouped_properties.pop(description_property, [])
            if isinstance(description, Literal)
        ]

        try:
            literal = choices[0]
        except IndexError:
            return None

        literal_value = literal.value

        match literal_value:
            case str() as string:
                return string

            case minidom.Document() as xml_document:
                return Syntax(
                    xml_document.toxml(),
                    'xml',
                )

            case something_else:
                type_of_something_else = type(something_else)
                raise ValueError(
                    f'What is this? {something_else} '   # noqa: WPS326
                    f'is a {type_of_something_else}!',   # noqa: WPS326
                )

    @functools.cached_property
    def properties(self) -> Widget | None:
        """Render properties table."""
        if not self.grouped_properties:
            return Static('No properties found ☹')

        return PropertiesContainer(*self.rows)

    def show(self) -> Widget:
        """Render the content."""
        return VerticalScroll(
            PageTitle(self.this),
            Static(self.description or ''),
            self.properties,
        )


class InverseProperties(TextualDefaultFacet):
    """Inverse properties view."""

    query_file_name = 'inverse-properties.sparql'

    def show(self) -> Widget:
        """Render the content."""
        return VerticalScroll(
            PageTitle(self.this, extra='[i]& its inverse RDF properties[/i]'),
            Static(self.description or ''),
            self.properties,
        )



from pathlib import Path

from jinja2 import Environment, FileSystemLoader
from rdflib import URIRef

from iolanta.facets.facet import Facet
from iolanta.namespaces import DATATYPES, IOLANTA, RDF, RDFS


class MkDocsMaterialInsidersMarkdownFacet(Facet[str]):
    """Render rdfs:Datatype nodes as mkdocs-material-insiders-markdown."""
    
    META = Path(__file__).parent / 'data' / 'mkdocs_material_insiders_markdown.yamlld'
    """Render rdfs:Datatype nodes as mkdocs-material-insiders-markdown."""

    @property
    def _template_env(self) -> Environment:
        """Jinja2 template environment."""
        template_path = Path(__file__).parent / 'templates'
        return Environment(
            loader=FileSystemLoader(str(template_path)),
            autoescape=False,
        )

    def show(self) -> str:
        """Render the datatype as markdown."""
        # Get the label using title facet
        label = self.render(
            self.this,
            as_datatype=DATATYPES.title,
        )

        # Get the comment/description
        comment_rows = self.query(
            """
            SELECT ?comment WHERE {
                $this rdfs:comment ?comment .
            }
            LIMIT 1
            """,
            this=self.this,
        )
        comment = str(comment_rows[0]['comment']) if comment_rows else None

        # Get all types (rdf:type)
        type_rows = self.query(
            """
            SELECT ?type WHERE {
                $this rdf:type ?type .
            }
            """,
            this=self.this,
        )
        types = [
            {
                'uri': row['type'],
                'title': self.render(row['type'], as_datatype=DATATYPES.title),
            }
            for row in type_rows
        ]

        # Get all superclasses (rdfs:subClassOf)
        superclass_rows = self.query(
            """
            SELECT ?superclass WHERE {
                $this rdfs:subClassOf ?superclass .
            }
            """,
            this=self.this,
        )
        superclasses = [
            {
                'uri': row['superclass'],
                'title': self.render(row['superclass'], as_datatype=DATATYPES.title),
            }
            for row in superclass_rows
        ]

        template = self._template_env.get_template('datatype.jinja2.md')
        return template.render(
            label=label,
            comment=comment,
            types=types,
            superclasses=superclasses,
        )




from pathlib import Path

import funcy
from docutils.nodes import row
from rdflib import Literal, Namespace
from rich.table import Table

from iolanta import Facet
from iolanta.namespaces import DATATYPES, LOCAL

LEXINFO = Namespace('https://www.lexinfo.net/ontology/2.0/lexinfo#')


class RichDeclensionTable(Facet[Table]):
    """Declension forms of something."""

    META = Path(__file__).parent / 'data' / 'declension.yamlld'

    def show(self) -> Table:
        """Render declension forms of something."""
        rows = self.stored_query('declension.sparql', this=self.this)
        form_by_number_and_person = [
            ((row['number'], row['person']), self.render(row['declension'], as_datatype=DATATYPES.title))
            for row in rows
        ]

        declensions = funcy.group_values(form_by_number_and_person)

        table = Table('', 'Singular', 'Plural', title='Declension')

        table_rows = [
            [
                ', '.join(declensions[number, person])
                for number in [LEXINFO.singular, LEXINFO.plural]
            ]
            for person in [LEXINFO.firstPerson, LEXINFO.secondPerson, LEXINFO.thirdPerson]
        ]

        row_titles = ['1st', '2nd', '3rd']

        for title, table_row in zip(row_titles, table_rows):
            table.add_row(title, *table_row)

        return table
